# Copyright (C) 2006-2020 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import ast
import logging
import os
import subprocess
import sys
import time


def run_cmd(args, success_return_code=(0,), logargs=None, stdin=None, run_background=False):
    logargs = logargs or args
    stdoutdata = ''
    if not run_background:
        proc = subprocess.Popen(args, stdin=stdin,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE, 
                                close_fds=True)
        stdoutdata, stderrdata = proc.communicate()
        if proc.returncode not in success_return_code:
            raise Exception('Command: %s failed (%s): stderr: %s stdout: %s'
                            % (logargs, proc.returncode, stderrdata, stdoutdata))
    else:
        stdoutdata = subprocess.Popen(args, stdin=stdin,
                                      stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return stdoutdata


def readfile(filepath):
    with open(filepath) as f:
        return f.read()


def writefile(filepath, content):
    with open(filepath, 'w') as f:
        f.write(content)
        return f.close()


def get_all_vms():
    """Get all VMs through xm."""
    vms = []
    lines = run_cmd(['xm', 'list']).splitlines()[1:]
    for line in lines:
        vm_name = line.split()[0].strip()
        if vm_name != 'Domain-0':
            vms.append(vm_name)
    return vms


def get_all_vm_vif_macs():
    """Return a vif-mac table of all VMs in a list."""
    vms = get_all_vms()
    vif_mac_table = {}
    for vm in vms:
        lines = run_cmd(['xm', 'network-list', vm]).splitlines()[1:]
        for line in lines:
            network_info = line.split()
            node_name = 'backend' + network_info[8].strip().split('backend')[1]
            mac = network_info[2].strip()
            vif_mac_table[node_name] = mac
    return vif_mac_table


def get_bond_active_slave(bond):
    """Get an active-backup bond's current active slave."""
    return readfile('/sys/class/net/%s/bonding/active_slave' % bond).strip()


def get_active_backup_bonds():
    """Get all bonds that are in active-backup mode."""
    bonds = readfile('/sys/class/net/bonding_masters').strip().split()
    active_backup_bonds = []
    for bond in bonds:
        active_slave = get_bond_active_slave(bond)
        if bond != 'bond0' and active_slave != '' and \
        readfile('/sys/class/net/%s/bonding/mode' % bond).strip().split()[1] == '1':
            active_backup_bonds.append(bond)
    return active_backup_bonds


def get_bond_bridges_and_vlans(bond):
    """Get all bridges and vlans that are created on a bond."""
    upper_dirs = run_cmd(['find', '/sys/class/net/%s/' % bond, '-name', 'upper_*']).splitlines()
    bridges = []
    vlans = []
    for upper_dir in upper_dirs:
        upper_dir = upper_dir.strip()
        device_name = upper_dir.split('upper_', 1)[1]
        if os.path.isdir(upper_dir + '/bridge'):
            bridges.append(device_name)
        else:
            vlans.append(device_name)
    return bridges, vlans


def get_bridge_macs(br, bond, vm_vif_mac_table):
    """Look up all MAC addresss attaching to a bridge using the existing vif-mac table."""
    macs = []
    vif_dirs = run_cmd(['find','/sys/class/net/%s/' % br, '-name', 'lower_*']).splitlines()
    for vif_dir in vif_dirs:
        vif_dir = vif_dir.strip()
        vif_name = vif_dir.split('lower_', 1)[1]
        if vif_name != bond:
            if os.path.exists('%s/device/nodename' % vif_dir):
                nodename = readfile('%s/device/nodename' % vif_dir).strip()
                if nodename in vm_vif_mac_table:
                    mac = vm_vif_mac_table[nodename]
                    macs.append(mac)
    return macs


def init_arp_reverse_table():
    """Initiate the global arp_reverse_table from show-endpoints"""
    global arp_reverse_table
    arp_reverse_table = {}
    # init arp_reverse_table with show-endpoints provided by customers
    if os.path.exists(endpoints_table_file):
        lines = readfile(endpoints_table_file).splitlines()
        for line in lines:
            info = line.strip().split()
            if len(info) > 1 and ':' in info[0] and '.' in info[1]:
                mac = info[0].lower()
                ip = info[1]
                if mac not in arp_reverse_table:
                    arp_reverse_table[mac] = [ip]
                elif len(arp_reverse_table[mac]) < max_ip_per_mac and ip not in arp_reverse_table[mac]:
                    arp_reverse_table[mac].append(ip)
                elif len(arp_reverse_table[mac]) == max_ip_per_mac:
                    logger.warning("%s cannot map to %s for the MAC's IP number has reached its limit: %s"
                                   % (ip, mac, arp_reverse_table[mac]))
    else:
        logger.warning("Unable to find %s, current MAC-IP table acquired is %s"
                       % (endpoints_table_file, arp_reverse_table))
    logger.debug("arp_reverse_table after updated by %s: %s" % (endpoints_table_file, arp_reverse_table))


def validate_ip(ip):
    addr_nums = ip.split('.')
    if len(addr_nums) != 4:
        return False
    for addr_num in addr_nums:
        if not addr_num.isdigit():
            return False
        if int(addr_num) < 0 or int(addr_num) > 255:
            return False
    return True


def send_frame_with_ip_and_mac(ip, mac, interface, count=3):
    """Use the arping tool to send out frames with fake IP & MAC address through specific interface."""
    if validate_ip(ip):
        cmd = ['/opt/ovs-extra/arping', '-PU', ip, '-S', ip, '-ps', mac, '-i', interface, '-c', str(count)]
        run_cmd(cmd, success_return_code=(0,1), run_background=True)
        logger.debug("ARPing started on behalf of VNIC IP/MAC: %s/%s ... with packet count of %s" 
                     % (ip, mac, count))
    else:
        logger.warning("Invalid IP %s for %s, please check show-endpoints" % (ip, mac))


def check_unsent_mac_pool():
    """Check MACs in unsent_mac_pool to see if their IPs are acquired and send the arping frames if so."""
    global unsent_mac_pool
    mac_sent = []
    for mac in unsent_mac_pool:
        if mac in arp_reverse_table:
            ips = arp_reverse_table[mac]
            br = unsent_mac_pool[mac][0]
            frame_count = unsent_mac_pool[mac][1]
            for ip in ips:
                send_frame_with_ip_and_mac(ip, mac, br, frame_count)
            mac_sent.append(mac)
            logger.info("Arping frames with MAC address %s in the unsent pool is now sent with IPs %s"
                        % (mac, ips))
    for mac in mac_sent:
        del unsent_mac_pool[mac]
    logger.debug("Current unsent_mac_pool is: %s" % unsent_mac_pool)


def bond_slave_port_update(frame_count):
    """
       Store all bonds and their active slaves in a table and monitor every active slave change.
       If any bond active slave changes, first find all the bridges attaching to it and its vlans.
       And then find all the vif MAC addresses that are attaching to those bridges.
       Finally, update the network switch by sending frames with these source MAC addresses
       and update the bond slave table for the next time detection.
    """
    global bond_slave_table
    global unsent_mac_pool
   
    init_arp_reverse_table() 
    check_unsent_mac_pool()
    bonds = get_active_backup_bonds()

    for bond in bonds:
        active_slave = get_bond_active_slave(bond)

        if bond not in bond_slave_table:
            bond_slave_table[bond] = active_slave

        elif active_slave != bond_slave_table[bond]:
 
            logger.info("bond '%s' active slave changed from '%s' to '%s'" 
                        % (bond, bond_slave_table[bond], active_slave)) 
            brs, vlans = get_bond_bridges_and_vlans(bond)
            logger.info("bridges attached to bond '%s' are: %s" % (bond, brs))
            logger.info("vlans attached to bond '%s' are: %s" % (bond, vlans))
            vm_vif_mac_table = get_all_vm_vif_macs()

            for br in brs:
                macs = get_bridge_macs(br, bond, vm_vif_mac_table)
                logger.info("mac addresses attached to bridge '%s' are: %s" % (br, macs))
    
                for mac in macs:
                    if mac in arp_reverse_table:
                        ips = arp_reverse_table[mac]
                        for ip in ips:
                            send_frame_with_ip_and_mac(ip, mac, br, frame_count)
                    else:
                        unsent_mac_pool[mac] = [br, frame_count]
                        logger.warning("IP not found for MAC: %s, packet will be sent once IP acquired" % mac)

            for vlan in vlans:
                vlan_brs = get_bond_bridges_and_vlans(vlan)[0]
                logger.info("bridges attached to vlan %s are: %s" % (vlan, vlan_brs))
                for vlan_br in vlan_brs:
                    macs = get_bridge_macs(vlan_br, vlan, vm_vif_mac_table)
                    logger.info("mac addresses attached to vlan bridge '%s' are: %s" % (vlan_br, macs))

                    for mac in macs:
                        if mac in arp_reverse_table:
                            ips = arp_reverse_table[mac]
                            for ip in ips:
                                send_frame_with_ip_and_mac(ip, mac, vlan_br, frame_count)
                        else:
                            unsent_mac_pool[mac] = [vlan_br, frame_count]
                            logger.warning("IP not found for MAC: %s, packet will be sent once IP acquired" % mac)

            bond_slave_table[bond] = active_slave
    logger.debug("Current bond_slave_table: %s" % bond_slave_table)


service = 'ovs-arping'
bond_slave_table = {}
arp_reverse_table = {}
unsent_mac_pool = {}
endpoints_table_file = '/opt/ovs-extra/show-endpoints'
max_ip_per_mac = 200
frame_count = 3		# 3 frames are sent for a single affacted mac address by default

logger = logging.getLogger('%s' % service)
logger.setLevel(logging.DEBUG)
handler = logging.FileHandler('/var/log/%s.log' % service)
handler.setLevel(logging.INFO)
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s', '%Y-%m-%d %H:%M:%S'))
logger.addHandler(handler)

if __name__ == '__main__':

    try:
        logger.info("%s service started.\n" % service)
        if len(sys.argv) > 1 and sys.argv[1] == 'debug':
            handler.setLevel(logging.DEBUG)
        while True:
            bond_slave_port_update(frame_count)
            # every frame takes 1 second to be sent so 'frame_count' seconds are required for one VNIC
            # the update frames for all endpoint  are processed in parallel
            # so adding 1 second for all the rest of the process to complete before another round check
            time.sleep(frame_count + 1)
    except Exception as e:
        logger.error("ERROR: %s, please check the issue and restart %s service\n" % (str(e), service))
